package POEx::Clock;

use 5.008008;
use strict;
use warnings;

use POE;
use POE::Component::Generic;
use Storable qw( dclone );
use POSIX qw( SIGTERM SIGALRM );


our $VERSION = '0.0100';

#######################################
sub spawn
{
    my( $package, $args ) = @_;
    my $self = $package->new( $args );
    $self->init;
    POE::Session->create(
            object_states => [
                $self => [ qw( _start _stop shutdown skew child_error child_exit ) ]
            ] );
    return;
}

#######################################
sub debug
{
    my $self = shift;
    return unless $self->{debug};
    print STDERR "$self->{alias}: ", @_, "\n";
}

#######################################
sub new
{
    my( $self, $args ) = @_;
    return bless { %$args, args=>dclone $args }, $self;
}

#######################################
sub init
{
    my( $self ) = @_;
    $self->{alias} ||= 'clock-skew';
    $self->{debug}  ||= 0;
}

#######################################
sub _start
{
    my( $self ) = @_[OBJECT, ARG0..$#_];
    $self->debug( '_start' );
    $poe_kernel->alias_set( $self->{alias} );
    $poe_kernel->sig( shutdown => 'shutdown' );
    $self->build_child;
    $self->query_child;
}

#######################################
sub _stop
{
    my( $self ) = @_[OBJECT, ARG0..$#_];
    $self->debug( '_stop' );
}

#######################################
sub shutdown
{
    my( $self ) = @_[OBJECT, ARG0..$#_];
    $self->{shutdown} = 1;
    $self->debug( 'shutdown' );
    $self->kill_child;
    $poe_kernel->alias_remove( $self->{alias} );
}

#######################################
sub build_child
{
    my( $self ) = @_;
    return if $self->{child} or $self->{shutdown};
    $self->{detecting} = 0;
    $self->{child} = POE::Component::Generic->spawn(
                                package => 'POEx::Clock::Child',
                                object_options => [ $self->{args} ],
                                verbose => $self->{verbose},
                                # debug   => $self->{debug},
                                error   => 'child_error',
                                on_exit => 'child_exit',
                                methods => [ qw( init detect_skew ) ]
                            );
    $self->{child}->init({});
}

#######################################
sub rebuild_child
{
    my( $self ) = @_;
    $self->kill_child;
    $self->build_child;
    $self->query_child;
}

#######################################
sub kill_child
{
    my( $self ) = @_;
    return unless $self->{child};
    $self->{child}->kill( SIGTERM );
    delete $self->{child};
}

#######################################
sub query_child
{
    my( $self ) = @_;
    return if $self->{detecting};
    $self->{detecting} = 1;
    $self->{child}->detect_skew( { event => 'skew', wantarray => 1 } );
}

#######################################
sub skew
{
    my( $self, $hash, $epsilon, $ex ) = @_[ OBJECT, ARG0..$#_ ];
    my $expected = $ex->[0] + $ex->[1] / 1_000_000;
    $self->debug( join ' ', 'skew', $epsilon, $expected );
    $self->{detecting} = 0;
    $poe_kernel->signal( $poe_kernel, 'clock-skew', $epsilon, $expected );
    $self->query_child;
}

#######################################
sub child_error
{
    my( $self, $err ) = @_[ OBJECT, ARG0 ];
    if( $err->{stderr} ) {
        $self->debug( $err->{stderr} );
    }
    elsif( ($err->{operation} eq 'read' or $err->{operation} eq 'write') and not $err->{errnum} ) {
        $self->debug( 'child closed' );
        # $self->rebuild_child;
    }
    else {
        die "child $err->{operation} error: ($err->{errnum}) $err->{errstr}";
    }
}

sub child_exit
{
    my( $self, $obj ) = @_[ OBJECT, ARG0 ];
    $self->debug( 'child exit' );
    $self->rebuild_child;
}

##########################################################################
package POEx::Clock::Child;

use POSIX::RT::Timer;
use POSIX;
use Time::HiRes qw( gettimeofday tv_interval );

#######################################
sub new
{
    my( $package, $args ) = @_;
    my $self = bless $args, $package;
    return $self;
}

#######################################
sub init
{
    my( $self ) = @_;
    $self->{interval} ||= 5*60;        # 5 minutes;
    $self->{interval} = int $self->{interval};
    $self->{precision} ||= 1;
    $self->{signal} ||= SIGALRM;
    return;    
}

#######################################
sub detect_skew
{
    my( $self ) = @_;
    $self->set_sig;
    $self->build_timer;
    while( 1 ) {
        if( $self->{triggered} ) {                  
            # If the parent didn't send another request back fast enough
            # we could have missed a signal.  This is an attempt to compensate
            $self->{now} += $self->{interval}*$self->{triggered};
            $self->{triggered} = 0;
        }
        my $expected = $self->calc_expected;
        pause() while not $self->{triggered};
        $self->now;
        $self->{triggered} = 0;
        my $epsilon = $self->calc_epsilon( $expected );
        my $ep = abs( $epsilon );
        return $epsilon, $expected if $ep > $self->{precision};
    }
}

#######################################
sub now
{
    my( $self ) = @_;
    $self->{now} = [ gettimeofday ];
}

#######################################
sub calc_expected
{
    my( $self ) = @_;
    return [ $self->{now}[0] + $self->{interval}, $self->{now}[1] ];
}

#######################################
sub calc_epsilon
{
    my( $self, $expected ) = @_;
    my $epsilon = tv_interval( $expected, [gettimeofday] );
    $self->{debug} and warn "$$: expected=$expected->[0] epsilon=$epsilon\n";
    return $epsilon;
}

#######################################
sub set_sig
{
    my( $self ) = @_;
    return if $self->{sigact};

    $self->{triggered} = 0;
    my $handler = sub { 
            $self->{debug} and warn "$$: Triggered\n";
            $self->{triggered}++;
        };
    $self->{flags} = 0;

    $self->{sigset} = POSIX::SigSet->new( $self->{signal} );
    $self->{sigact} = POSIX::SigAction->new( $handler, $self->{sigset}, $self->{flags} );
    sigaction( $self->{signal}, $self->{sigact} );
}

#######################################
sub build_timer
{
    my( $self ) = @_;
    return if $self->{timer};
    $self->{timer} = POSIX::RT::Timer->new( 
                               value    => $self->{interval},
                               interval => $self->{interval},
                               signal   => $self->{signal}
                             );
    $self->now;
}

1;

__END__

=head1 NAME

POEx::Clock - Detect clock skew

=head1 SYNOPSIS

  use POEx::Clock;

    POEx::Clock->spawn({ interval => $seconds,
                         precision => 0.5,
                      });

    $poe_kernel->sig( 'clock-skew' => 'skewed' );


    sub skewed_handler
    {
        my( $self, $epsilon, $expected_time ) = @_[ OBJECT, ARG0, ARG1 ];
        # The system time has been skewed by $epsilon seconds
        # You might want to adjust your kernel timers
    }        


=head1 DESCRIPTION

POEx::Clock implements a phase-locked loop that is used to detect jumps in
the system clock.  The phase-locked loop is a monotonic timer created with
L<POSIX::RT::Timer>.  This timer triggers a signal.  When the signal is
received, the current time is check against the expected time.  Any
discrepancy is due to system clock skew.

=head1 METHODS

=head2 spawn

=over 4

=item interval

=item precision

=item signal

=item alias

=back

=head1 SIGNALS

This component emits the following signals.

=head2 clock-skew

    $poe_kernel->sig( 'clock-skew' => 'skewed' );

    sub skewed_handler
    {
        my( $self, $epsilon, $expected_time ) = @_[ OBJECT, ARG0, ARG1 ];
    }

When 


=head1 SEE ALSO

L<POE>, L<POSIX::RT::Timer>.

=head1 AUTHOR

Philip Gwyn, E<lt>gwyn-at-cpan.orgE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2012 by Philip Gwyn

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.8 or,
at your option, any later version of Perl 5 you may have available.


=cut
